from .registry import (Registry, Serialiser)
from .reasonable import (Reasonable, SerReasonableObject)
from ..interface import (PromisedObject)
from ..utility import (object_name, look_up, importable)
from ..workflow import (Workflow, NodeData, FunctionNode, ArgumentAddress,
                        ArgumentKind, reset_workflow, get_workflow)
from ..storable import (Storable)
# from .as_dict import (AsDict)
# from enum import Enum
from inspect import isfunction, ismethod
# from collections import namedtuple
from itertools import count
# import json
# import sys


class SerDict(Serialiser):
    def __init__(self):
        super(SerDict, self).__init__(dict)

    def encode(self, obj, make_rec):
        return make_rec(dict(obj))

    def decode(self, cls, data):
        return cls(data)


class SerTuple(Serialiser):
    """Tuples get converted to lists during serialisation.
    We want to get tuples back, so make this explicit."""
    def __init__(self):
        super(SerTuple, self).__init__(tuple)

    def encode(self, obj, make_rec):
        return make_rec(list(obj))

    def decode(self, cls, data):
        return cls(data)


class SerEnum(Serialiser):
    def __init__(self, cls):
        super(SerEnum, self).__init__(cls)

    def encode(self, obj, make_rec):
        return make_rec(obj.name)

    def decode(self, cls, data):
        return cls[data]


class SerNamedTuple(Serialiser):
    def __init__(self, cls):
        super(SerNamedTuple, self).__init__(cls)

    def encode(self, obj, make_rec):
        return make_rec(dict(obj._asdict()))

    def decode(self, cls, data):
        return cls(**data)


def _remap_links(remap, links):
    return [{'node': remap[source],
             'to': [{'node': remap[node],
                     'address': address}
                    for node, address in target]}
            for source, target in links.items()]


class SerWorkflow(Serialiser):
    def __init__(self):
        super(SerWorkflow, self).__init__(Workflow)

    def encode(self, obj, make_rec):
        remap = dict(zip(obj.nodes.keys(), count()))
        return make_rec({'root': remap[obj.root],
                         'nodes': list(obj.nodes.values()),
                         'links': _remap_links(remap, obj.links)})

    def decode(self, cls, data):
        root = data['root']

        nodes = dict(zip(count(), data['nodes']))

        links = {l['node']: {(target['node'], target['address'])
                             for target in l['to']}
                 for l in data['links']}

        return reset_workflow(Workflow(root, nodes, links))


class SerPromisedObject(Serialiser):
    def __init__(self):
        super(SerPromisedObject, self).__init__(PromisedObject)

    def encode(self, obj, make_rec):
        return make_rec({'workflow': get_workflow(obj)})

    def decode(self, cls, data):
        return PromisedObject(data['workflow'])


class SerMethod(Serialiser):
    def __init__(self):
        super(SerMethod, self).__init__('<method>')

    def encode(self, obj, make_rec):
        return make_rec({'class': object_name(obj.__member_of__),
                         'method': obj.__name__})

    def decode(self, cls, data):
        cls = look_up(data['class'])
        return getattr(cls, data['method'])


class SerBoundMethod(Serialiser):
    def __init__(self):
        super(SerBoundMethod, self).__init__('<boundmethod>')

    def encode(self, obj, make_rec):
        return make_rec({
            'self': obj.__self__,
            'name': obj.__name__})

    def decode(self, _, data):
        return getattr(data['self'], data['name'])


class SerImportable(Serialiser):
    def __init__(self):
        super(SerImportable, self).__init__('<importable>')

    def encode(self, obj, make_rec):
        return make_rec(object_name(obj))

    def decode(self, cls, data):
        return look_up(data)


class SerStorable(Serialiser):
    def __init__(self, cls):
        super(SerStorable, self).__init__(cls)

    def encode(self, obj, make_reca):
        return make_reca(
            {'type': object_name(type(obj)),
             'dict': obj.as_dict()})

    def decode(self, _, data):
        cls = look_up(data['type'])
        return cls.from_dict(**data['dict'])


class SerNode(Serialiser):
    def __init__(self):
        super(SerNode, self).__init__(FunctionNode)

    def encode(self, obj, make_rec):
        return make_rec(dict(obj.data._asdict()))

    def decode(self, cls, data):
        return FunctionNode.from_node_data(NodeData(**data))


def _noodles_hook(obj):
    if '__member_of__' in dir(obj) and obj.__member_of__:
        return '<method>'

    if importable(obj):
        return '<importable>'

    if ismethod(obj):
        return '<boundmethod>'

    if isfunction(obj):
        return '<importable>'

    return None


def registry():
    """Returns the Noodles base serialisation registry."""
    return Registry(
        types={
            dict: SerDict(),
            tuple: SerTuple(),
            Reasonable: SerReasonableObject(Reasonable),
            ArgumentKind: SerEnum(ArgumentKind),
            FunctionNode: SerNode(),
            ArgumentAddress: SerNamedTuple(ArgumentAddress),
            Workflow: SerWorkflow(),
            Storable: SerStorable(Storable),
            PromisedObject: SerPromisedObject()
        },
        hooks={
            '<method>': SerMethod(),
            '<boundmethod>': SerBoundMethod(),
            '<importable>': SerImportable(),
        },
        hook_fn=_noodles_hook,
        default=Serialiser(object),
    )



from .decorator import (
    PromisedObject, schedule, schedule_hint, has_scheduled_methods,
    unwrap, update_hints)
from .functions import (
    delay, gather, lift, unpack, quote, unquote, find_first)
from .annotated_value import (AnnotatedValue)
from .exceptions import (JobException)

__all__ = ['delay', 'gather', 'schedule_hint', 'schedule', 'unpack',
           'has_scheduled_methods', 'unwrap', 'update_hints', 'AnnotatedValue',
           'JobException', 'lift', 'PromisedObject', 'quote', 'unquote',
           'find_first']



from . import schedule, schedule_hint


@schedule
def add(x, y):
    return x + y


@schedule_hint(display="{a} + {b}", confirm=True)
def log_add(a, b):
    return a + b


@schedule
def sub(x, y):
    return x - y


@schedule
def mul(x, y):
    return x*y


@schedule
def accumulate(lst, start=0):
    return sum(lst, start)



from nose.plugins.skip import SkipTest

try:
    from noodles.prov import prov_key
    from noodles.run.run_with_prov import (
        run_single, run_parallel, run_parallel_opt)
except ImportError as e:
    raise SkipTest(str(e))

from noodles import serial, gather, schedule_hint
from noodles.tutorial import (add, mul, sub, accumulate)

from noodles.run.worker import run_job
from noodles.run.job_keeper import JobKeeper

import os


def test_prov_00():
    reg = serial.base()
    a = add(3, 4)
    b = sub(3, 4)
    c = add(3, 4)
    d = add(4, 3)

    enc = [reg.deep_encode(x._workflow.root_node) for x in [a, b, c, d]]
    key = [prov_key(o) for o in enc]
    assert key[0] == key[2]
    assert key[1] != key[0]
    assert key[3] != key[0]


def test_prov_01():
    reg = serial.base()
    a = add(3, 4)

    enc = reg.deep_encode(a._workflow.root_node)
    dec = reg.deep_decode(enc)

    result = run_job(0, dec)
    assert result.value == 7


def test_prov_02():
    db_file = "prov1.json"

    A = add(1, 1)
    B = sub(3, A)

    multiples = [mul(add(i, B), A) for i in range(6)]
    C = accumulate(gather(*multiples))

    result = run_single(C, serial.base, db_file)
    assert result == 42
    os.unlink(db_file)


def test_prov_03():
    db_file = "prov2.json"

    A = add(1, 1)
    B = sub(3, A)

    multiples = [mul(add(i, B), A) for i in range(6)]
    C = accumulate(gather(*multiples))

    result = run_parallel(C, 4, serial.base, db_file, JobKeeper(keep=True))
    assert result == 42
    os.unlink(db_file)


@schedule_hint(store=True)
def add2(x, y):
    return x + y


def test_prov_04():
    db_file = "prov3.json"

    A = add2(1, 1)
    B = sub(3, A)

    multiples = [mul(add2(i, B), A) for i in range(6)]
    C = accumulate(gather(*multiples))

    result = run_parallel_opt(C, 4, serial.base, db_file)
    assert result == 42
    os.unlink(db_file)


@schedule_hint(store=True)
def fib(n):
    if n < 2:
        return 1
    else:
        return add(fib(n - 1), fib(n - 2))


def test_prov_05():
    import time
    db_file = "testjobs.json"

    wf = fib(20)
    start = time.time()
    result = run_parallel(wf, 4, serial.base, db_file)
    end = time.time()

    assert (end - start) < 5.0  # weak test
    assert result == 10946

    os.unlink(db_file)



from noodles import (schedule, run_single, gather, lift)
from noodles.tutorial import (add, sub, mul)


class A:
    pass


@schedule
def f(a):
    return a.x + a.y


def test_lift_01():
    a = A()
    a.x = add(1, 2)
    a.y = sub(9, 11)
    b = f(lift(a))

    result = run_single(b)
    assert result == 1




from .pretty_term import OutStream
import sys


class DumbDisplay:
    """Monochrome, dumb term display"""
    def __init__(self, error_filter=None):
        self.jobs = {}
        self.out = OutStream(sys.stdout)
        self.errors = []
        self.error_filter = error_filter

    def start(self, key, job, _):
        if job.hints and 'display' in job.hints:
            self.add_job(key, job.hints)
            self.out << job.hints['display'].format(
                **job.bound_args.arguments) << "\n"

    def done(self, key, data, msg):
        pass

    def error(self, key, _, msg):
        pass

    def add_job(self, key, hints):
        for k in self.jobs:
            self.jobs[k]['line'] += 1
        self.jobs[key] = {'line': 1}
        self.jobs[key].update(hints)

    def error_handler(self, job, xcptn):
        self.errors.append((job, xcptn))

    def report(self):
        if len(self.errors) == 0:
            self.out << "+---(success)\n"

        else:
            self.out << "+---(ERROR!)\n\n"

            for job, e in self.errors:
                msg = 'ERROR '
                if 'display' in job.hints:
                    msg += job.hints['display'].format(
                        **job.bound_args.arguments)
                else:
                    msg += 'calling {} with {}'.format(
                        job.foo.__name__, dict(job.bound_args.arguments)
                    )

                print(msg)
                err_msg = self.error_filter(e)
                if err_msg:
                    print(err_msg)
                else:
                    print(e)

    def __call__(self, key, status, data, err):
        # self.q = q
        # for status, key, data in q.source():
        getattr(self, status)(key, data, err)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # self.wait()

        if exc_type:
            if exc_type is KeyboardInterrupt:
                self.out << "\n" << "User interrupt detected, abnormal exit.\n"
                return True

            print("Internal error encountered. Contact the developers.")
            return False

        self.report()

    def wait(self):
        self.q.wait()



import noodles
import nose


@nose.tools.raises(TypeError)
def test_iterate():
    a = noodles.delay((1,2,3))
    b, c, d = a
    e = noodles.gather(d, c, b)

    result = noodles.run_single(e)
    assert result == (3, 2, 1)



from noodles import run_process, schedule, serial


@schedule
def writes_to_stdout():
    print("Hello Noodles!")
    return 42


def test_capture_output():
    a = writes_to_stdout()
    result = run_process(a, n_processes=1, registry=serial.base, use_msgpack=True)
    assert result == 42
